// tiny modification to calculate timer2 reload values for different MCU clock speeds - see the lines below including F_CPU, July 2022 ceptimus
// Phil_G's lockdown project, May 2020.
// A simple four channel RC encoder for the NRF24L01 with proper FHSS hopping, one packet every 5ms
// Uses sixteen RF channels well spread through the band and the system works well.
// Has single-channel escapement emulation mix, Vtail, Elevons, minute-minder, remote failsafe, throttle lock etc
// expo, rates, +-400 range + trim,  two S/C keying speeds, dual (+/-) stick mapping, minute-minder, etc
// Please note that the calibration procedure is different to the previous version, see below.
// Please do not use Arduino bootloader for programming, its startup delay messes things up
// Connections:
// Stick pots are wired between ground and regulated 5v from Arduino. Wipers connected as follows:
// A0=aileron, A1=elevator, A2=rudder, A3=throttle, A6=expo, A7=rates. Always use this wiring order regardless of channel order.
// Ground A6 and A7 with links or Spektrum bind-plugs if rates & expo are not required. Dont leave them floating.
// Unwanted digital functions can be omitted except for S/C button on D8 which is needed for calibrating the sticks.
// Note that the DIY More ProMini Strong has annotations for D0 and D1 reversed - the pins actually number 1023Fradio4567.
// D0=spare, D1=spare, D2=buzzer, D3=increment minute timer, D4=set failsafe button, D5=50:50 mixer, D6=75:25 mixer,
// D7=throttle lock, D8=compound s/c, D9 & D10=NRF24, D11, D12 & D13=NRF24 SPI. Possibly A4 & A5 for OLED display?
// Features:
// Calibration - center trims, hold either s/c button in, switch on, still holding button move all controls to extremes including auxilliary pots
// Center all controls including throttle and aux pots. Release button.  At this point the throttle-lock warning should chirp because the throttle is centred.
// Do not switch off during calibration, the calibration values cant be saved if the set has been switched off!
// Throttle response can optionally be made non-linear during calibration by 'centering' the throttle stick to say 2/3 rather than to neutral stick.
// Variable expo & rates on Futaba ch1, 2 & 4, ditto JR ch2, 3 & 4. Elevon mixer on Futaba ch1 & ch2, JR ch2 & 3, 75% aileron, 25% elevator or 50:50 for V-tails
// Mixer option can only be changed on power-up (avoids accidentally switching during flight).  Single-channel emulation mix, compound only.
// Servo reversing by holding sticks over on power up (saved to eeprom).
// Auto and manual throttle-lock with warning.
// Follow the story on the single-channel forum http://mode-zero.uk/viewtopic.php?f=27&t=844
// Diagram and documentation on http://www.singlechannel.co.uk see Archive page
// If you use this software in whole or in part please acknowledge its source - Phil_G on most forums, philg@talk21.com  http://www.singlechannel.co.uk
// Special thanks to the late Dave Burton and Mike Kitchen for their improved expo maths.

#include <SPI.h>
#include <nRF24L01.h>
#include <RF24.h>

// timer 2 OCR2A values (1024 prescaler) for 5ms and 6ms interrupt intervals
#define TOP_5ms (((F_CPU >> 10) + 100) / 200)
#define TOP_6ms (((F_CPU >> 10) + 83) / 167)

#define full_n -400 // full negative deflection of sticks
#define full_p +400 // full positive deflection of sticks
#define neutralPulse 1500
#define buzzer    2  // inactivity buzzer pin
#define min_tmr   3  // set/clear minute minder button pin
#define set_fs    4  // set F/S from transmitter
#define mix_50    5  // mixer 50:50 pin
#define mix_75    6  // mixer 75:25 pin
#define sw_lock   7  // throttle-lock toggle pin
#define sw_btn    8  // pushbutton for s/c compound pin & stick calibration
#define inactFrames 30000  // 10 minutes

/************* Radio config - make it unique! **********************************/
const byte pipeOut[] = "12345";   // last 5 digits of your phone number is sufficiently unique in our small group
const byte mychannels[] = {38, 8, 68, 78, 43, 28, 53, 3, 23, 48, 33, 58, 63, 13, 73, 18}; // choose 16 RANDOM but non-adjacent channels 1 to 82.  Don't use these!
/*******************************************************************************/

bool sc_fast = 0;  // optionally set to 1 for fast rattler (Elmic Compact)

int rfchan = 0; // NRF channel index

RF24 radio(10, 9);  // RF-NANO and new builds, ce=10, csn=9.   Phils original was RF24 radio(9, 10) dont use this for new builds except LGT8F RF-NANO

struct Packet { //    *** note NRF24L01 has a max 32 byte packet size ***
  unsigned int aileron  = 1500;
  unsigned int elevator = 1500;
  unsigned int throttle = 1100;
  unsigned int rudder   = 1500;
};

Packet data;

int raw, flips, ch, calibrated = 1, stickcalHi[] = {0, 0, 0, 0}, stickcalLo[] = {1023, 1023, 1023, 1023}, stickcalMid[] = {512, 512, 512, 512};
int a0_val, chtemp, contrl_pos[] = {1500, 1500, 1500, 1500}, min_minder = 0;
volatile int framecounter = 0, sync = 0;
long int expo, rates, input, expoCalc1, expoCalc2, expoCalc3;
byte reverse[] = {0, 0, 0, 0, 0, 0};
static int rotor28[] = { -36, -72, -108, -144, -180, -216, -252, -216, -180, -144, -108, -72, -36,  0,  36,  72, 108, 144, 180, 216, 252, 216, 180, 144, 108, 72, 36, 0};
static int rotor24[] = { -42, -84, -126, -168, -210, -252, -210, -168, -126,  -84,  -42,   0,  42, 84, 126, 168, 210, 252, 210, 168, 126,  84,  36,   0,   0,  0,  0, 0};
byte tlock = 1, mix_option = 0, rotorptr = 0, timer_flg = 0;
float scrudder = 0, scelev = 0;
unsigned int inact = 0, fiftyhz = 0, tmr_held = 0;
unsigned long startTime, mins = 0;

#include <EEPROM.h>

void setup()  {
  pinMode(buzzer, OUTPUT);
  pinMode(mix_50, INPUT_PULLUP);
  pinMode(mix_75, INPUT_PULLUP);
  pinMode(sw_lock, INPUT_PULLUP);
  pinMode(sw_btn, INPUT_PULLUP);
  pinMode(set_fs, INPUT_PULLUP);
  pinMode(min_tmr, INPUT_PULLUP);
  pinMode(set_fs, INPUT_PULLUP);

  while (digitalRead(sw_btn) == 0) {  // calibrate sticks
    calibrated = 0;
    for (int stick = 0; stick < 4; ++stick) {
      raw = analogRead(stick);
      if (raw > stickcalHi[stick]) stickcalHi[stick] = raw;
      if (raw < stickcalLo[stick]) stickcalLo[stick] = raw;
      stickcalMid[stick] = raw;  // save neutral sticks as button is released
    }
  }  // calibrate button released

  if (calibrated == 0) {
    for (ch = 0; ch < 4; ch++) {
      EEPROMWriteInt(ch * 6, stickcalLo[ch]);      // eeprom locations  0,  6, 12, 18  (decimal)
      EEPROMWriteInt(ch * 6 + 2, stickcalMid[ch]); // eeprom locations  2,  8, 14, 20  (decimal)
      EEPROMWriteInt(ch * 6 + 4, stickcalHi[ch]);  // eeprom locations  4, 10, 16, 22  (decimal)
    }
    calibrated = 1;
    digitalWrite(buzzer, HIGH);
    delay(30);
    digitalWrite(buzzer, LOW);
  }

  for (ch = 0; ch < 4; ch++) {
    stickcalLo[ch]  = EEPROMReadInt(ch * 6);       // eeprom locations  0,  6, 12, 18  (decimal)
    stickcalMid[ch] = EEPROMReadInt(ch * 6 + 2);   // eeprom locations  2,  8, 14, 20  (decimal)
    stickcalHi[ch]  = EEPROMReadInt(ch * 6 + 4);   // eeprom locations  4, 10, 16, 22  (decimal)
    reverse[ch] = EEPROM.read(ch + 24) & 1;        // eeprom locations 24, 25, 26, 27 (decimal)
    reverse[3] = 0;                                // no throttle reverse for safety
  }

  if (digitalRead(mix_50) == 0) mix_option = 1;
  if (digitalRead(mix_75) == 0) mix_option = 2;
 
  // check for reversing, stick over on power-up, not throttle though.
  for (ch = 0; ch < 3; ch++) {
    contrl_pos[ch] = map(analogRead(ch), stickcalLo[ch], stickcalHi[ch], full_n, full_p);
    if (contrl_pos[ch] > full_p - 50 || contrl_pos[ch] < full_n + 50) {
      reverse[ch] ^= B00000001;
      EEPROM.write(24 + ch, reverse[ch]);
    }
  }

  // Tx config
  radio.begin(); delayMicroseconds(500);
  radio.openWritingPipe(pipeOut); delayMicroseconds(500);
  radio.setAutoAck(false); delayMicroseconds(500);
  radio.setDataRate(RF24_250KBPS); delayMicroseconds(500);
  radio.setPALevel(RF24_PA_MAX); delayMicroseconds(500); // RF24_PA_MIN, RF24_PA_LOW, RF24_PA_HIGH and RF24_PA_MAX
  radio.setChannel(mychannels[0]); delayMicroseconds(500);
  radio.stopListening(); // transmit mode

  // set up 5ms packet send interrupt, CTC mode, prescaler to /1024
  TCCR2A = 0;        // reset Timer2 control reg A
  TCCR2A |= (1 << WGM21);     // set WGM21 bit, CTC mode
  TCCR2B |= (1 << CS22) | (1 << CS21) | (1 << CS20); // 1024 prescaler
  OCR2A = TOP_5ms;           // top = 78 with 16 MHz crystal, but adjusts to suit processor clock speed
  TIMSK2 |= (1 << OCIE2A);    // enable timer2 compare interrupt
  sei();        // enable global interrupts

  for (ch = 0; ch < 3; ++ch) {
    digitalWrite(buzzer, HIGH);
    delay(30);
    digitalWrite(buzzer, LOW);
    delay(30);
  }

}

void loop() {

  // read sticks
  for (ch = 0; ch < 4; ch++) {
    raw = analogRead(ch);
    if (raw > stickcalMid[ch]) contrl_pos[ch] = map (raw, stickcalMid[ch], stickcalHi[ch], 0, full_p);
    else contrl_pos[ch] = map (raw, stickcalLo[ch], stickcalMid[ch], full_n, 0);
  }

  // Optionally, and only if you fully understand the array allocations, you can set any unused analogues here instead of
  // paralleling unused analogue inputs.  Either method prevents floating inputs 'wandering' or following an adjacent pin.
  // Physical wiring is: A0=aileron stick pot, A1=elevator, A2=rudder, A3=throttle, A4=aux1, A5=aux2, and
  // A0 is your 'primary steering' control - on a rudder/elevator/throttle set, the rudder servo is plugged into the aileron channel. SOP.
  // Array numbering is: contrl_pos[0]=aileron stick pot, contrl_pos[1]=elevator, contrl_pos[2]=rudder, contrl_pos[3]=throttle,
  // So for example on a 3ch set, zero the rudder element (contrl_pos[0] is primary steering) like this:
  // contrl_pos[2]=0;  // example for a 3ch rudder-elevator-throttle set
  // contrl_pos[2]=0; contrl_pos[3]=0; // example for a 2ch aileron-elevator set

  // single-channel emulation
  scrudder = 0;

  // compound button
  if (rotorptr == 0 && digitalRead(sw_btn) == 0) ++rotorptr; // if esc is at rest and we press the button, start the rotation
  if (rotorptr != 0) {
    scelev = 0;

    if (sc_fast == 1) {
      scrudder = rotor24[rotorptr];
      if (digitalRead(sw_btn) == 0 && (rotorptr == 5 || rotorptr == 17)) --rotorptr; // do pawl stops @ 90 and 270 degrees
      if (digitalRead(sw_btn) == 0 && rotorptr == 22) {
        --rotorptr;  // do pawl stops @ 355 degrees kue
        scelev = 250;
      }
      ++rotorptr; if (rotorptr == 24) rotorptr = 0;
    }
    else {
      scrudder = rotor28[rotorptr];
      if (digitalRead(sw_btn) == 0 && (rotorptr == 6 || rotorptr == 20)) --rotorptr; // do pawl stops @ 90 and 270 degrees
      if (digitalRead(sw_btn) == 0 && rotorptr == 26) {
        --rotorptr;  // do pawl stops @ 355 degrees kue
        scelev = 250;
      }
      ++rotorptr; if (rotorptr == 28) rotorptr = 0;
    }
  }

  contrl_pos[0] = contrl_pos[0] + scrudder;
  contrl_pos[1] = contrl_pos[1] + scelev;

  // throttle lock/hold-off
  if (digitalRead(sw_lock) == 0) tlock = 1;
  if (digitalRead(sw_lock) == 1 && contrl_pos[3] < (full_n + 60)) tlock = 0;
  if (contrl_pos[3] > (full_n + 60) && tlock) {
    if (bitRead(fiftyhz, 0) && bitRead(fiftyhz, 5) && bitRead(fiftyhz, 3)) digitalWrite(buzzer, HIGH); else digitalWrite(buzzer, LOW);
  }
  if (tlock) contrl_pos[3] = full_n;

  a0_val = contrl_pos[0];

  /*Rates and Expo.   With thanks to Mike Kitchen and Prof Dave Burton RIP = Phil
     We apply expo first and then the rates
     Original Expo Formula by Prof Dave Burton and coded by Mike Kitchen
     The expo calculation is not "true" exponential but a cubic polynominal approximating an exponential function
     The basic form of the expo equation is: output = ((1 - expo) * input) + (expo * input^3)
     The formula can be re-written as: output = input - (expo * input) + (expo * input^3)
     where  -1 <= output  =< 1, 0 <= expo =< 1 and -1<= input =< 1.
     But we need -500 <= output =< 500, 0 <= expo =< 100 and -500 <= input =< 500 (ie the input and output in the range +/-500 and expo in the range 0-100)
     If we re-arrange for input and output +/-500 and expo 0 to 100 then we get the following formula
     output = input - (expo * input /100) + (expo * input * input * input)/25,000,000
     Expo is set by the expo pot in the range 0-1023. Map this to 0% to 100%.

     The rates formula is output = input * pot value / 1023   (hence need inverting to match Phil_G rates+
     where input and output are in the range +/-500 and pot value in the range 0-1023
  */

  expo = (long) map(analogRead(A6), 0, 1023, 0, 100); //read the expo pot and map into the range 0 to 100
  rates = (long) 1023 - analogRead(A7); //read the rates pot and leave in the range 0 - 1023
  if (rates < 100) rates = 100;
  for (ch = 0; ch < 3; ch++) { // for aileron, elevator and rudder
    input = (long int) contrl_pos[ch];
    expoCalc1 = input * expo / 100; //Do the first element of the calculation
    expoCalc2 = (input * input * input) / 100L; //Do the second element of the calculation in two stages to ensure it doesn't overflow "long" variable ie divide by 100 and then a further 250,000
    expoCalc2 = expoCalc2 * expo / 158000L;  // 250000L for +-500.  Ths value is adjusted for +-400.
    expoCalc3 = input - expoCalc1 + expoCalc2;  //Sum the three terms of the expo calc
    contrl_pos[ch] = (int) (expoCalc3 * rates / 1023); // the rates calculation
  }

  /*End of new expo and rates */

  // format the frame
  for (ch = 0; ch < 4; ch++) {
    contrl_pos[ch] += neutralPulse;
    contrl_pos[ch] = constrain(contrl_pos[ch], 900, 2100);
    if (reverse[ch] == 1) contrl_pos[ch] = 3000 - contrl_pos[ch];
  }

  // check if mixer switched in
  if (mix_option == 1) {
    contrl_pos[0] /= 2;
    contrl_pos[1] /= 2;
    chtemp = contrl_pos[1];
    contrl_pos[1] = contrl_pos[0] + 1500 - contrl_pos[1];
    contrl_pos[0] += chtemp;
  }

  if (mix_option == 2) {
    contrl_pos[0] *= 3;
    contrl_pos[0] /= 4;
    contrl_pos[1] /= 4;
    chtemp = contrl_pos[1];
    contrl_pos[1] = contrl_pos[0] + 750 - contrl_pos[1];
    contrl_pos[0] += chtemp;
  }

  // Futaba or Spektrum/JR channel order is set at the receiver
  data.aileron     = contrl_pos[0];
  data.elevator    = contrl_pos[1];
  data.rudder      = contrl_pos[2];
  data.throttle    = contrl_pos[3];

  // set failsafe?
  if (digitalRead(set_fs) == 0) data.aileron += 2000;

  // hold button to set minute-minder
  if (digitalRead(min_tmr) == 0) { // minute minder button pressed
    if (++min_minder == 30) {
      timer_flg = 1;
      digitalWrite(buzzer, HIGH);
      delay(30);
      digitalWrite(buzzer, LOW);
      min_minder = 0; // button pressed time
      startTime = millis();
      mins += 60000L; // 60000 ms per minute
    }
  }
  else min_minder = 0;

  // countdown minute-minder
  if (digitalRead(min_tmr) == 1 && timer_flg == 1) {
    if (millis() > startTime + mins) {
      if (mins > 0) {
        mins = 0;  //  ensures we start with sound not the pause between
        fiftyhz = 0x01c1;
      }
      if (bitRead(fiftyhz, 0) == 1 && bitRead(fiftyhz, 6) == 1 && bitRead(fiftyhz, 7) == 1 && bitRead(fiftyhz, 8) == 1) digitalWrite(buzzer, HIGH);
      else digitalWrite(buzzer, LOW);
    }
  }

  // quick dab of the button clears minute-minder
  if (digitalRead(min_tmr) == 0 && timer_flg == 1  && mins == 0 && min_minder < 30) {
    timer_flg = 0; // minute minder button pressed to cancel alarm
    mins = 0;
    digitalWrite(buzzer, LOW);
  }

  if (inact < inactFrames && timer_flg == 0 && tlock == 0) digitalWrite(buzzer, LOW);

  // inactivity timer. 10 mins is approx 30000 frames.
  if (a0_val < -100 || a0_val > 100) inact = 0;
  if (digitalRead(sw_btn) == 0) inact = 0;
  if (++inact > inactFrames) {
    if (bitRead(fiftyhz, 2) == 1 && bitRead(fiftyhz, 5) == 1) digitalWrite(buzzer, HIGH); else digitalWrite(buzzer, LOW);
  }
  if (bitRead(fiftyhz, 0) == 0) digitalWrite(buzzer, LOW);

  while (sync == 0); // wait for whatever remains of 20ms (four packets)
  sync = 0;

  ++fiftyhz;
}

// This function will write a 2 byte integer to the eeprom at the specified address and address + 1
void EEPROMWriteInt(int p_address, int p_value)
{
  byte lowByte = p_value % 256;
  byte highByte = p_value / 256;
  EEPROM.write(p_address, lowByte);
  EEPROM.write(p_address + 1, highByte);
}

//This function will read a 2 byte integer from the eeprom at the specified address and address + 1
unsigned int EEPROMReadInt(int p_address)
{
  byte lowByte = EEPROM.read(p_address);
  byte highByte = EEPROM.read(p_address + 1);
  return lowByte + highByte * 256;
}

ISR(TIMER2_COMPA_vect) {   // send a packet every 5ms
  radio.write(&data, sizeof(Packet));
  rfchan = (rfchan + 1) % 16;
  radio.setChannel(mychannels[rfchan]);
  framecounter = (framecounter + 1) % 4; // 20ms
  if (framecounter == 3) sync = 1;
}

// If you use this software in whole or in part please acknowledge its source - Phil_G on most forums, philg@talk21.com  http://www.singlechannel.co.uk
// Read the story here: http://mode-zero.uk/viewtopic.php?f=42&t=971    and its predecessor here: http://mode-zero.uk/viewtopic.php?f=42&t=945
